

/////////////////////////////////////////////
//      Solar Position Algorithm (SPA)     //
//                   for                   //
//        Solar Radiation Application      //
//                                         //
//               May 12, 2003              //
//                                         //
//   Filename: SPA.C                       //
//                                         //
//   Afshin Michael Andreas                //
//   Afshin.Andreas@NREL.gov (303)384-6383 //
//                                         //
//   Measurement & Instrumentation Team    //
//   Solar Radiation Research Laboratory   //
//   National Renewable Energy Laboratory  //
//   1617 Cole Blvd, Golden, CO 80401      //
/////////////////////////////////////////////

/////////////////////////////////////////////
//   See the SPA.H header file for usage   //
//                                         //
//   This code is based on the NREL        //
//   technical report "Solar Position      //
//   Algorithm for Solar Radiation         //
//   Application" by I. Reda & A. Andreas  //
/////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
//
//   NOTICE
//   Copyright (C) 2008-2011 Alliance for Sustainable Energy, LLC, All Rights Reserved
//
//The Solar Position Algorithm ("Software") is code in development prepared by employees of the
//Alliance for Sustainable Energy, LLC, (hereinafter the "Contractor"), under Contract No.
//DE-AC36-08GO28308 ("Contract") with the U.S. Department of Energy (the "DOE"). The United
//States Government has been granted for itself and others acting on its behalf a paid-up, non-
//exclusive, irrevocable, worldwide license in the Software to reproduce, prepare derivative
//works, and perform publicly and display publicly. Beginning five (5) years after the date
//permission to assert copyright is obtained from the DOE, and subject to any subsequent five
//(5) year renewals, the United States Government is granted for itself and others acting on
//its behalf a paid-up, non-exclusive, irrevocable, worldwide license in the Software to
//reproduce, prepare derivative works, distribute copies to the public, perform publicly and
//display publicly, and to permit others to do so. If the Contractor ceases to make this
//computer software available, it may be obtained from DOE's Office of Scientific and Technical
//Information's Energy Science and Technology Software Center (ESTSC) at P.O. Box 1020, Oak
//Ridge, TN 37831-1020. THIS SOFTWARE IS PROVIDED BY THE CONTRACTOR "AS IS" AND ANY EXPRESS OR
//IMPLIED WARRANTIES, INCLUDING BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
//AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE CONTRACTOR OR THE
//U.S. GOVERNMENT BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
//WHATSOEVER, INCLUDING BUT NOT LIMITED TO CLAIMS ASSOCIATED WITH THE LOSS OF DATA OR PROFITS,
//WHICH MAY RESULT FROM AN ACTION IN CONTRACT, NEGLIGENCE OR OTHER TORTIOUS CLAIM THAT ARISES
//OUT OF OR IN CONNECTION WITH THE ACCESS, USE OR PERFORMANCE OF THIS SOFTWARE.
//
//The Software is being provided for internal, noncommercial purposes only and shall not be
//re-distributed. Please contact Jennifer Ramsey (Jennifer.Ramsey@nrel.gov) in the NREL
//Commercialization and Technology Transfer Office for information concerning a commercial
//license to use the Software.
//
//As a condition of using the Software in an application, the developer of the application
//agrees to reference the use of the Software and make this Notice readily accessible to any
//end-user in a Help|About screen or equivalent manner.
//
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
// Revised 27-FEB-2004 Andreas
//         Added bounds check on inputs and return value for spa_calculate().
// Revised 10-MAY-2004 Andreas
//         Changed temperature bound check minimum from -273.15 to -273 degrees C.
// Revised 17-JUN-2004 Andreas
//         Corrected a problem that caused a bogus sunrise/set/transit on the equinox.
// Revised 18-JUN-2004 Andreas
//         Added a "function" input variable that allows the selecting of desired outputs.
// Revised 21-JUN-2004 Andreas
//         Added 3 new intermediate output values to SPA structure (srha, ssha, & sta).
// Revised 23-JUN-2004 Andreas
//         Enumerations for "function" were renamed and 2 were added.
//         Prevented bound checks on inputs that are not used (based on function).
// Revised 01-SEP-2004 Andreas
//         Changed a local variable from integer to double.
// Revised 12-JUL-2005 Andreas
//         Put a limit on the EOT calculation, so that the result is between -20 and 20.
// Revised 26-OCT-2005 Andreas
//         Set the atmos. refraction correction to zero, when sun is below horizon.
//         Made atmos_refract input a requirement for all "functions".
//         Changed atmos_refract bound check from +/- 10 to +/- 5 degrees.
// Revised 07-NOV-2006 Andreas
//         Corrected 3 earth periodic terms in the L_TERMS array.
//         Corrected 2 earth periodic terms in the R_TERMS array.
// Revised 10-NOV-2006 Andreas
//         Corrected a constant used to calculate topocentric sun declination.
//         Put a limit on observer hour angle, so result is between 0 and 360.
// Revised 13-NOV-2006 Andreas
//         Corrected calculation of topocentric sun declination.
//         Converted all floating point inputs in spa structure to doubles.
// Revised 27-FEB-2007 Andreas
//         Minor correction made as to when atmos. refraction correction is set to zero.
// Revised 21-JAN-2008 Andreas
//         Minor change to two variable declarations.
// Revised 12-JAN-2009 Andreas
//         Changed timezone bound check from +/-12 to +/-18 hours.
// Revised 14-JAN-2009 Andreas
//         Corrected a constant used to calculate ecliptic mean obliquity.
// Revised 01-APR-2013 Andreas
//         Replace floor with new integer function for tech. report consistency, no affect on results.
//         Add "utility" function prototypes to header file for use with NREL's SAMPA.
//         Rename 4 "utility" function names (remove "sun") for clarity with NREL's SAMPA.
//         Added delta_ut1 as required input, which the fractional second difference between UT and UTC.
//         Time must be input w/o delta_ut1 adjustment, instead of assuming adjustment was pre-applied.
// Revised 10-JUL-2014 Andreas
//         Change second in spa_data structure from an integer to double to allow fractional second
// Revised 08-SEP-2014 Andreas
//         Corrected description of azm_rotation in header file
//         Limited azimuth180 to range of 0 to 360 deg (instead of -180 to 180) for tech report consistency
//         Changed all variables names from azimuth180 to azimuth_astro
//         Renamed 2 "utility" function names for consistency
///////////////////////////////////////////////////////////////////////////////////////////////////

/* The original code was modified heavily by Val Gretchev March 2015 to run on an ARM microcontroller */

#include <math.h>
#include "spa.h"

#include "USB_Driver.h"
#include "Buffers.h"



#define PI 3.141592653589L
#define SUN_RADIUS 0.26667L

#define L_COUNT 6
#define B_COUNT 2
#define R_COUNT 5
#define Y_COUNT 63

#define L_MAX_SUBCOUNT 64
#define B_MAX_SUBCOUNT 5
#define R_MAX_SUBCOUNT 40

enum {TERM_A, TERM_B, TERM_C, TERM_COUNT};
enum {TERM_X0, TERM_X1, TERM_X2, TERM_X3, TERM_X4, TERM_X_COUNT};
enum {TERM_PSI_A, TERM_PSI_B, TERM_EPS_C, TERM_EPS_D, TERM_PE_COUNT};
enum {JD_MINUS, JD_ZERO, JD_PLUS, JD_COUNT};
enum {SUN_TRANSIT, SUN_RISE, SUN_SET, SUN_COUNT};

#define TERM_Y_COUNT TERM_X_COUNT

const int l_subcount[L_COUNT] = {64,34,20,7,3,1};
const int b_subcount[B_COUNT] = {5,2};
const int r_subcount[R_COUNT] = {40,10,6,2,1};









///////////////////////////////////////////////////////////////////////////////////////////////////
///  Earth Periodic Terms
///////////////////////////////////////////////////////////////////////////////////////////////////
const long double L_TERMS[L_COUNT][L_MAX_SUBCOUNT][TERM_COUNT] =
{
    {
        {175347046.0L,0,0},
        {3341656.0L,4.6692568L,6283.07585L},
        {34894.0L,4.6261L,12566.1517L},
        {3497.0L,2.7441L,5753.3849L},
        {3418.0L,2.8289L,3.5231L},
        {3136.0L,3.6277L,77713.7715L},
        {2676.0L,4.4181L,7860.4194L},
        {2343.0L,6.1352L,3930.2097L},
        {1324.0L,0.7425L,11506.7698L},
        {1273.0L,2.0371L,529.691L},
        {1199.0L,1.1096L,1577.3435L},
        {990.0L,5.233L,5884.927L},
        {902.0L,2.045L,26.298L},
        {857.0L,3.508L,398.149L},
        {780.0L,1.179L,5223.694L},
        {753.0L,2.533L,5507.553L},
        {505.0L,4.583L,18849.228L},
        {492.0L,4.205L,775.523L},
        {357.0L,2.92L,0.067L},
        {317.0L,5.849L,11790.629L},
        {284.0L,1.899L,796.298L},
        {271.0L,0.315L,10977.079L},
        {243.0L,0.345L,5486.778L},
        {206.0L,4.806L,2544.314L},
        {205.0L,1.869L,5573.143L},
        {202.0L,2.458L,6069.777L},
        {156.0L,0.833L,213.299L},
        {132.0L,3.411L,2942.463L},
        {126.0L,1.083L,20.775L},
        {115.0L,0.645L,0.98L},
        {103.0L,0.636L,4694.003L},
        {102.0L,0.976L,15720.839L},
        {102.0L,4.267L,7.114L},
        {99.0L,6.21L,2146.17L},
        {98.0L,0.68L,155.42L},
        {86.0L,5.98L,161000.69L},
        {85.0L,1.3L,6275.96L},
        {85.0L,3.67L,71430.7L},
        {80.0L,1.81L,17260.15L},
        {79.0L,3.04L,12036.46L},
        {75.0L,1.76L,5088.63L},
        {74.0L,3.5L,3154.69L},
        {74.0L,4.68L,801.82L},
        {70.0L,0.83L,9437.76L},
        {62.0L,3.98L,8827.39L},
        {61.0L,1.82L,7084.9L},
        {57.0L,2.78L,6286.6L},
        {56.0L,4.39L,14143.5L},
        {56.0L,3.47L,6279.55L},
        {52.0L,0.19L,12139.55L},
        {52.0L,1.33L,1748.02L},
        {51.0L,0.28L,5856.48L},
        {49.0L,0.49L,1194.45L},
        {41.0L,5.37L,8429.24L},
        {41.0L,2.4L,19651.05L},
        {39.0L,6.17L,10447.39L},
        {37.0L,6.04L,10213.29L},
        {37.0L,2.57L,1059.38L},
        {36.0L,1.71L,2352.87L},
        {36.0L,1.78L,6812.77L},
        {33.0L,0.59L,17789.85L},
        {30.0L,0.44L,83996.85L},
        {30.0L,2.74L,1349.87L},
        {25.0L,3.16L,4690.48L}
    },
    {
        {628331966747.0L,0,0},
        {206059.0L,2.678235L,6283.07585L},
        {4303.0L,2.6351L,12566.1517L},
        {425.0L,1.59L,3.523L},
        {119.0L,5.796L,26.298L},
        {109.0L,2.966L,1577.344L},
        {93.0L,2.59L,18849.23L},
        {72.0L,1.14L,529.69L},
        {68.0L,1.87L,398.15L},
        {67.0L,4.41L,5507.55L},
        {59.0L,2.89L,5223.69L},
        {56.0L,2.17L,155.42L},
        {45.0L,0.4L,796.3L},
        {36.0L,0.47L,775.52L},
        {29.0L,2.65L,7.11L},
        {21.0L,5.34L,0.98L},
        {19.0L,1.85L,5486.78L},
        {19.0L,4.97L,213.3L},
        {17.0L,2.99L,6275.96L},
        {16.0L,0.03L,2544.31L},
        {16.0L,1.43L,2146.17L},
        {15.0L,1.21L,10977.08L},
        {12.0L,2.83L,1748.02L},
        {12.0L,3.26L,5088.63L},
        {12.0L,5.27L,1194.45L},
        {12.0L,2.08L,4694.0L},
        {11.0L,0.77L,553.57L},
        {10.0L,1.3L,6286.6L},
        {10.0L,4.24L,1349.87L},
        {9.0L,2.7L,242.73L},
        {9.0L,5.64L,951.72L},
        {8.0L,5.3L,2352.87L},
        {6.0L,2.65L,9437.76L},
        {6.0L,4.67L,4690.48L}
    },
    {
        {52919.0L,0,0},
        {8720.0L,1.0721L,6283.0758L},
        {309.0L,0.867L,12566.152L},
        {27.0L,0.05L,3.52L},
        {16.0L,5.19L,26.3L},
        {16.0L,3.68L,155.42L},
        {10.0L,0.76L,18849.23L},
        {9.0L,2.06L,77713.77L},
        {7.0L,0.83L,775.52L},
        {5.0L,4.66L,1577.34L},
        {4.0L,1.03L,7.11L},
        {4.0L,3.44L,5573.14L},
        {3.0L,5.14L,796.3L},
        {3.0L,6.05L,5507.55L},
        {3.0L,1.19L,242.73L},
        {3.0L,6.12L,529.69L},
        {3.0L,0.31L,398.15L},
        {3.0L,2.28L,553.57L},
        {2.0L,4.38L,5223.69L},
        {2.0L,3.75L,0.98L}
    },
    {
        {289.0L,5.844L,6283.076L},
        {35.0L,0,0},
        {17.0L,5.49L,12566.15L},
        {3.0L,5.2L,155.42L},
        {1.0L,4.72L,3.52L},
        {1.0L,5.3L,18849.23L},
        {1.0L,5.97L,242.73L}
    },
    {
        {114.0L,3.142L,0},
        {8.0L,4.13L,6283.08L},
        {1.0L,3.84L,12566.15L}
    },
    {
        {1.0L,3.14L,0}
    }
};

const long double B_TERMS[B_COUNT][B_MAX_SUBCOUNT][TERM_COUNT] =
{
    {
        {280.0L,3.199L,84334.662L},
        {102.0L,5.422L,5507.553L},
        {80.0L,3.88L,5223.69L},
        {44.0L,3.7L,2352.87L},
        {32.0L,4.0L,1577.34L}
    },
    {
        {9.0L,3.9L,5507.55L},
        {6.0L,1.73L,5223.69L}
    }
};

const long double R_TERMS[R_COUNT][R_MAX_SUBCOUNT][TERM_COUNT] =
{
    {
        {100013989.0L,0,0},
        {1670700.0L,3.0984635L,6283.07585L},
        {13956.0L,3.05525L,12566.1517L},
        {3084.0L,5.1985L,77713.7715L},
        {1628.0L,1.1739L,5753.3849L},
        {1576.0L,2.8469L,7860.4194L},
        {925.0L,5.453L,11506.77L},
        {542.0L,4.564L,3930.21L},
        {472.0L,3.661L,5884.927L},
        {346.0L,0.964L,5507.553L},
        {329.0L,5.9L,5223.694L},
        {307.0L,0.299L,5573.143L},
        {243.0L,4.273L,11790.629L},
        {212.0L,5.847L,1577.344L},
        {186.0L,5.022L,10977.079L},
        {175.0L,3.012L,18849.228L},
        {110.0L,5.055L,5486.778L},
        {98.0L,0.89L,6069.78L},
        {86.0L,5.69L,15720.84L},
        {86.0L,1.27L,161000.69L},
        {65.0L,0.27L,17260.15L},
        {63.0L,0.92L,529.69L},
        {57.0L,2.01L,83996.85L},
        {56.0L,5.24L,71430.7L},
        {49.0L,3.25L,2544.31L},
        {47.0L,2.58L,775.52L},
        {45.0L,5.54L,9437.76L},
        {43.0L,6.01L,6275.96L},
        {39.0L,5.36L,4694.0L},
        {38.0L,2.39L,8827.39L},
        {37.0L,0.83L,19651.05L},
        {37.0L,4.9L,12139.55L},
        {36.0L,1.67L,12036.46L},
        {35.0L,1.84L,2942.46L},
        {33.0L,0.24L,7084.9L},
        {32.0L,0.18L,5088.63L},
        {32.0L,1.78L,398.15L},
        {28.0L,1.21L,6286.6L},
        {28.0L,1.9L,6279.55L},
        {26.0L,4.59L,10447.39L}
    },
    {
        {103019.0L,1.10749L,6283.07585L},
        {1721.0L,1.0644L,12566.1517L},
        {702.0L,3.142L,0},
        {32.0L,1.02L,18849.23L},
        {31.0L,2.84L,5507.55L},
        {25.0L,1.32L,5223.69L},
        {18.0L,1.42L,1577.34L},
        {10.0L,5.91L,10977.08L},
        {9.0L,1.42L,6275.96L},
        {9.0L,0.27L,5486.78L}
    },
    {
        {4359.0L,5.7846L,6283.0758L},
        {124.0L,5.579L,12566.152L},
        {12.0L,3.14L,0},
        {9.0L,3.63L,77713.77L},
        {6.0L,1.87L,5573.14L},
        {3.0L,5.47L,18849.23L}
    },
    {
        {145.0L,4.273L,6283.076L},
        {7.0L,3.92L,12566.15L}
    },
    {
        {4.0L,2.56L,6283.08L}
    }
};

///////////////////////////////////////////////////////////////////////////////////////////////////
///  Periodic Terms for the nutation in longitude and obliquity
///////////////////////////////////////////////////////////////////////////////////////////////////

const int Y_TERMS[Y_COUNT][TERM_Y_COUNT]=
{
    {0,0,0,0,1},
    {-2,0,0,2,2},
    {0,0,0,2,2},
    {0,0,0,0,2},
    {0,1,0,0,0},
    {0,0,1,0,0},
    {-2,1,0,2,2},
    {0,0,0,2,1},
    {0,0,1,2,2},
    {-2,-1,0,2,2},
    {-2,0,1,0,0},
    {-2,0,0,2,1},
    {0,0,-1,2,2},
    {2,0,0,0,0},
    {0,0,1,0,1},
    {2,0,-1,2,2},
    {0,0,-1,0,1},
    {0,0,1,2,1},
    {-2,0,2,0,0},
    {0,0,-2,2,1},
    {2,0,0,2,2},
    {0,0,2,2,2},
    {0,0,2,0,0},
    {-2,0,1,2,2},
    {0,0,0,2,0},
    {-2,0,0,2,0},
    {0,0,-1,2,1},
    {0,2,0,0,0},
    {2,0,-1,0,1},
    {-2,2,0,2,2},
    {0,1,0,0,1},
    {-2,0,1,0,1},
    {0,-1,0,0,1},
    {0,0,2,-2,0},
    {2,0,-1,2,1},
    {2,0,1,2,2},
    {0,1,0,2,2},
    {-2,1,1,0,0},
    {0,-1,0,2,2},
    {2,0,0,2,1},
    {2,0,1,0,0},
    {-2,0,2,2,2},
    {-2,0,1,2,1},
    {2,0,-2,0,1},
    {2,0,0,0,1},
    {0,-1,1,0,0},
    {-2,-1,0,2,1},
    {-2,0,0,0,1},
    {0,0,2,2,1},
    {-2,0,2,0,1},
    {-2,1,0,2,1},
    {0,0,1,-2,0},
    {-1,0,1,0,0},
    {-2,1,0,0,0},
    {1,0,0,0,0},
    {0,0,1,2,0},
    {0,0,-2,2,2},
    {-1,-1,1,0,0},
    {0,1,1,0,0},
    {0,-1,1,2,2},
    {2,-1,-1,2,2},
    {0,0,3,2,2},
    {2,-1,0,2,2},
};

const long double PE_TERMS[Y_COUNT][TERM_PE_COUNT]=
{
    {-171996.0L,-174.2L,92025.0L,8.9L},
    {-13187.0L,-1.6L,5736.0L,-3.1L},
    {-2274.0L,-0.2L,977.0L,-0.5L},
    {2062.0L,0.2L,-895.0L,0.5L},
    {1426.0L,-3.4L,54.0L,-0.1L},
    {712.0L,0.1,-7.0L,0},
    {-517.0L,1.2L,224.0L,-0.6L},
    {-386.0L,-0.4L,200.0L,0},
    {-301.0L,0,129.0L,-0.1L},
    {217.0L,-0.5L,-95.0L,0.3L},
    {-158.0L,0,0,0},
    {129.0L,0.1L,-70.0L,0},
    {123.0L,0,-53.0L,0},
    {63.0L,0,0,0},
    {63.0L,0.1L,-33.0L,0},
    {-59.0L,0,26.0L,0},
    {-58.0L,-0.1L,32.0L,0},
    {-51.0L,0,27.0L,0},
    {48.0L,0,0,0},
    {46.0L,0,-24.0L,0},
    {-38.0L,0,16.0L,0},
    {-31.0L,0,13.0L,0},
    {29.0L,0,0,0},
    {29.0L,0,-12.0L,0},
    {26.0L,0,0,0},
    {-22.0L,0,0,0},
    {21.0L,0,-10.0L,0},
    {17.0L,-0.1L,0,0},
    {16.0L,0,-8.0L,0},
    {-16.0L,0.1L,7.0L,0},
    {-15.0L,0,9.0L,0},
    {-13.0L,0,7.0L,0},
    {-12.0L,0,6.0L,0},
    {11.0L,0,0,0},
    {-10.0L,0,5.0L,0},
    {-8.0L,0,3.0L,0},
    {7.0L,0,-3.0L,0},
    {-7.0L,0,0,0},
    {-7.0L,0,3.0L,0},
    {-7.0L,0,3.0L,0},
    {6.0L,0,0,0},
    {6.0L,0,-3.0L,0},
    {6.0L,0,-3.0L,0},
    {-6.0L,0,3.0L,0},
    {-6.0L,0,3.0L,0},
    {5.0L,0,0,0},
    {-5.0L,0,3.0L,0},
    {-5.0L,0,3.0L,0},
    {-5.0L,0,3.0L,0},
    {4.0L,0,0,0},
    {4.0L,0,0,0},
    {4.0L,0,0,0},
    {-4.0L,0,0,0},
    {-4.0L,0,0,0},
    {-4.0L,0,0,0},
    {3.0L,0,0,0},
    {-3.0L,0,0,0},
    {-3.0L,0,0,0},
    {-3.0L,0,0,0},
    {-3.0L,0,0,0},
    {-3.0L,0,0,0},
    {-3.0L,0,0,0},
    {-3.0L,0,0,0},
};


/*****************************************************************************************************************************************/

long double rad2deg(long double radians)
{
    return (180.0L / PI) * radians;
}

long double deg2rad(long double degrees)
{
    return (PI / 180.0L) * degrees;
}

int integer(long double value)
{
    return (value);
}

long double limit_degrees(long double degrees)
{
    long double limited;

    degrees /= 360.0L;
    limited = 360.0L * (degrees - floor(degrees));
    if (limited < 0) limited += 360.0L;

    return limited;
}

long double limit_degrees180pm(long double degrees)
{
    long double limited;

    degrees /= 360.0L;
    limited = 360.0L * (degrees - floor(degrees));
    if      (limited < -180.0) limited += 360.0L;
    else if (limited >  180.0) limited -= 360.0L;

    return limited;
}

long double limit_degrees180(long double degrees)
{
    long double limited;

    degrees /= 180.0L;
    limited = 180.0L * (degrees - floor(degrees));
    if (limited < 0) limited += 180.0L;

    return limited;
}

long double limit_zero2one(long double value)
{
    long double limited;

    limited = value - floor(value);
    if (limited < 0) limited += 1.0L;

    return limited;
}

long double limit_minutes(long double minutes)
{
    long double limited = minutes;

    if      (limited < -20.0) limited += 1440.0L;
    else if (limited >  20.0) limited -= 1440.0L;

    return limited;
}

long double dayfrac_to_local_hr(long double dayfrac, long double timezone)
{
    return 24.0L * limit_zero2one(dayfrac + timezone / 24.0L);
}

long double third_order_polynomial(long double a, long double b, long double c, long double d, long double x)
{
    return ((a * x + b) * x + c) * x + d;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
int validate_inputs(spa_data *spa)
{
    if ((spa->year        < -2000) || (spa->year        > 6000)) return 1;
    if ((spa->month       < 1    ) || (spa->month       > 12  )) return 2;
    if ((spa->day         < 1    ) || (spa->day         > 31  )) return 3;
    if ((spa->hour        < 0    ) || (spa->hour        > 24  )) return 4;
    if ((spa->minute      < 0    ) || (spa->minute      > 59  )) return 5;
    if ((spa->second      < 0    ) || (spa->second      >=60  )) return 6;
    if ((spa->pressure    < 0    ) || (spa->pressure    > 5000)) return 12;
    if ((spa->temperature <= -273) || (spa->temperature > 6000)) return 13;
    if ((spa->delta_ut1   <= -1  ) || (spa->delta_ut1   >= 1  )) return 17;
    if ((spa->hour        == 24  ) && (spa->minute      > 0   )) return 5;
    if ((spa->hour        == 24  ) && (spa->second      > 0   )) return 6;

    if (fabs(spa->delta_t)       > 8000    ) return 7;
    if (fabs(spa->timezone)      > 18      ) return 8;
    if (fabs(spa->longitude)     > 180     ) return 9;
    if (fabs(spa->latitude)      > 90      ) return 10;
    if (fabs(spa->atmos_refract) > 5       ) return 16;
    if (spa->elevation           < -6500000) return 11;

    if ((spa->function == SPA_ZA_INC) || (spa->function == SPA_ALL))
    {
        if (fabs(spa->slope)         > 360) return 14;
        if (fabs(spa->azm_rotation)  > 360) return 15;
    }

    return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////
long double julian_day (int year, int month, int day, int hour, int minute, long double second,
                        long double dut1, long double tz)
{
    long double day_decimal, julian_day, a;

    day_decimal = day + (hour - tz + (minute + (second + dut1) / 60.0L) / 60.0L) / 24.0L;

    if (month < 3) {
        month += 12;
        year--;
    }

    julian_day = integer(365.25L * (year + 4716.0L)) + integer(30.6001L * (month + 1)) + day_decimal - 1524.5L;

    if (julian_day > 2299160.0L) {
        a = integer(year / 100L);
        julian_day += (2.0L - a + integer(a / 4.0L ));
    }

    return julian_day;
}

long double julian_century(long double jd)
{
    return (jd - 2451545.0L) / 36525.0L;
}

long double julian_ephemeris_day(long double jd, long double delta_t)
{
    return jd + delta_t / 86400.0L;
}

long double julian_ephemeris_century(long double jde)
{
    return (jde - 2451545.0L) / 36525.0L;
}

long double julian_ephemeris_millennium(long double jce)
{
    return (jce / 10.0L);
}


/*****************************************************************************************************************************************/


long double earth_periodic_term_summation(double terms[][TERM_COUNT], int count, long double jme){
    int i;
    long double sum = 0;

    for (i = 0; i < count; i++)
        sum += terms[i][TERM_A] * cos(terms[i][TERM_B] + terms[i][TERM_C] * jme);

    return sum;
}

long double earth_values(long double term_sum[], int count, long double jme){
    int i;
    long double sum = 0;

    for (i = 0; i < count; i++)
        sum += term_sum[i] * pow(jme, i);

    sum /= 1.0e8L;

    return sum;
}

long double earth_heliocentric_longitude(long double jme){
    long double sum[L_COUNT];
    int i;

    for (i = 0; i < L_COUNT; i++)
        sum[i] = earth_periodic_term_summation(L_TERMS[i], l_subcount[i], jme);

    return limit_degrees(rad2deg(earth_values(sum, L_COUNT, jme)));

}

long double earth_heliocentric_latitude(long double jme){
    long double sum[B_COUNT];
    int i;

    for (i = 0; i < B_COUNT; i++)
        sum[i] = earth_periodic_term_summation(B_TERMS[i], b_subcount[i], jme);

    return rad2deg(earth_values(sum, B_COUNT, jme));

}

long double earth_radius_vector(long double jme){
    long double sum[R_COUNT];
    int i;

    for (i = 0; i < R_COUNT; i++)
        sum[i] = earth_periodic_term_summation(R_TERMS[i], r_subcount[i], jme);

    return earth_values(sum, R_COUNT, jme);

}

/*****************************************************************************************************************************************/

long double geocentric_longitude(long double l)
{
    long double theta = l + 180.0L;

    if (theta >= 360) theta -= 360.0L;

    return theta;
}

long double geocentric_latitude(long double b)
{
    return -b;
}

long double mean_elongation_moon_sun(long double jce)
{
    return third_order_polynomial(1.0L / 189474.0L, -0.0019142L, 445267.11148L, 297.85036L, jce);
}

long double mean_anomaly_sun(long double jce)
{
    return third_order_polynomial(-1.0L / 300000.0L, -0.0001603L, 35999.05034L, 357.52772L, jce);
}

long double mean_anomaly_moon(long double jce)
{
    return third_order_polynomial(1.0L / 56250.0L, 0.0086972L, 477198.867398L, 134.96298L, jce);
}

long double argument_latitude_moon(long double jce)
{
    return third_order_polynomial(1.0L / 327270.0L, -0.0036825L, 483202.017538L, 93.27191L, jce);
}

long double ascending_longitude_moon(long double jce)
{
    return third_order_polynomial(1.0L / 450000.0L, 0.0020708L, -1934.136261L, 125.04452L, jce);
}

long double xy_term_summation(int i, long double x[TERM_X_COUNT])
{
    int j;
    long double sum=0;

    for (j = 0; j < TERM_Y_COUNT; j++)
        sum += x[j] * Y_TERMS[i][j];

    return sum;
}

/*****************************************************************************************************************************************/

void nutation_longitude_and_obliquity(long double jce, long double x[TERM_X_COUNT], long double *del_psi, long double *del_epsilon)
{
    int i;
    long double xy_term_sum, sum_psi=0, sum_epsilon=0;

    for (i = 0; i < Y_COUNT; i++) {
        xy_term_sum  = deg2rad(xy_term_summation(i, x));
        sum_psi += (PE_TERMS[i][TERM_PSI_A] + jce * PE_TERMS[i][TERM_PSI_B]) * sin(xy_term_sum);
        sum_epsilon += (PE_TERMS[i][TERM_EPS_C] + jce * PE_TERMS[i][TERM_EPS_D]) * cos(xy_term_sum);
    }

    *del_psi = sum_psi / 36000000.0L;
    *del_epsilon = sum_epsilon / 36000000.0L;
}

long double ecliptic_mean_obliquity(long double jme)
{
    long double u = jme / 10.0L;

    return 84381.448L + u * (-4680.93L + u * (-1.55L + u * (1999.25L + u * (-51.38L + u * (-249.67L +
                       u * (-39.05L + u * (7.12L + u * (27.87L + u * (5.79L + u * 2.45L)))))))));
}

long double ecliptic_true_obliquity(long double delta_epsilon, long double epsilon0)
{
    return delta_epsilon + epsilon0 / 3600.0L;
}

long double aberration_correction(long double r)
{
    return -20.4898L / (3600.0L * r);
}

long double apparent_sun_longitude(long double theta, long double delta_psi, long double delta_tau)
{
    return theta + delta_psi + delta_tau;
}

long double greenwich_mean_sidereal_time (long double jd, long double jc){
/*
    char FormatBuf[64];
    char* BufNum;

    BufNum = GetBuffer();
    LongDoubleToStr(jd, FormatBuf);
    sprintf(BufNum,"jd: ");
    strcat(BufNum,FormatBuf);
    USB_PutQueue(BufNum);
*/
    return limit_degrees(280.46061837L + 360.98564736629L *
                 (jd - 2451545.0L) + jc * jc * (0.000387933L - jc / 38710000.0L));
}

long double greenwich_sidereal_time (long double nu0, long double delta_psi, long double epsilon){
/*
    char FormatBuf[64];
    char* BufNum;

    BufNum = GetBuffer();
    LongDoubleToStr(epsilon, FormatBuf);
    sprintf(BufNum,"epsilon: ");
    strcat(BufNum,FormatBuf);
    USB_PutQueue(BufNum);
*/


    return nu0 + delta_psi * cos(deg2rad(epsilon));
}

long double geocentric_right_ascension(long double lamda, long double epsilon, long double beta)
{
    long double lamda_rad = deg2rad(lamda);
    long double epsilon_rad = deg2rad(epsilon);

    //char FormatBuf[64];
    //char* BufNum;

/*
    BufNum = GetBuffer();
    LongDoubleToStr(lamda, FormatBuf);
    sprintf(BufNum,"lamda: ");
    strcat(BufNum,FormatBuf);
    USB_PutQueue(BufNum);

    BufNum = GetBuffer();
    LongDoubleToStr(epsilon, FormatBuf);
    sprintf(BufNum,"epsilon: ");
    strcat(BufNum,FormatBuf);
    USB_PutQueue(BufNum);

    BufNum = GetBuffer();
    LongDoubleToStr(beta, FormatBuf);
    sprintf(BufNum,"beta: ");
    strcat(BufNum,FormatBuf);
    USB_PutQueue(BufNum);



    BufNum = GetBuffer();
    LongDoubleToStr(Temp1, FormatBuf);
    sprintf(BufNum,"Temp1: ");
    strcat(BufNum,FormatBuf);
    USB_PutQueue(BufNum);
*/
  //mikroC has a bug in atan2: the X and Y arguments have to be reversed
  //This is the original code
  //return limit_degrees(rad2deg  (atan2(sin(lamda_rad) * cos(epsilon_rad) - tan(deg2rad(beta)) * sin(epsilon_rad), cos(lamda_rad))));
    
  //This is the modified code
    return limit_degrees(rad2deg  (atan2(cos(lamda_rad), sin(lamda_rad) * cos(epsilon_rad) - tan(deg2rad(beta)) * sin(epsilon_rad))));
}

long double geocentric_declination(long double beta, long double epsilon, long double lamda)
{
    long double beta_rad = deg2rad(beta);
    long double epsilon_rad = deg2rad(epsilon);

    return rad2deg(asin(sin(beta_rad) * cos(epsilon_rad) +
                        cos(beta_rad) * sin(epsilon_rad) * sin(deg2rad(lamda))));
}

/*-----------------------------------------------------------------------------------------------*/
long double observer_hour_angle(long double nu, long double longitude, long double alpha_deg){
/*
    char FormatBuf[64];
    char* BufNum;
    long double Temp1;

    Temp1 = nu + longitude - alpha_deg;

    BufNum = GetBuffer();
    LongDoubleToStr(nu, FormatBuf);
    sprintf(BufNum,"nu: ");
    strcat(BufNum,FormatBuf);
    USB_PutQueue(BufNum);

    BufNum = GetBuffer();
    LongDoubleToStr(longitude, FormatBuf);
    sprintf(BufNum,"longitude: ");
    strcat(BufNum,FormatBuf);
    USB_PutQueue(BufNum);

    BufNum = GetBuffer();
    LongDoubleToStr(alpha_deg, FormatBuf);
    sprintf(BufNum,"alpha_deg: ");
    strcat(BufNum,FormatBuf);
    USB_PutQueue(BufNum);

    BufNum = GetBuffer();
    LongDoubleToStr(Temp1, FormatBuf);
    sprintf(BufNum,"Arithmetic result: ");
    strcat(BufNum,FormatBuf);
    USB_PutQueue(BufNum);
*/

    return limit_degrees(nu + longitude - alpha_deg);
}

long double sun_equatorial_horizontal_parallax(long double r)
{
    return 8.794L / (3600.0L * r);
}

/*****************************************************************************************************************************************/
void right_ascension_parallax_and_topocentric_dec(long double latitude, long double elevation, long double xi, long double h,
                                                  long double delta, long double *delta_alpha, long double *delta_prime)
{
    long double delta_alpha_rad;
    long double lat_rad = deg2rad(latitude);
    long double xi_rad = deg2rad(xi);
    long double h_rad = deg2rad(h);
    long double delta_rad = deg2rad(delta);
    long double u = atan(0.99664719L * tan(lat_rad));
    long double y = 0.99664719L * sin(u) + elevation * sin(lat_rad) / 6378140.0L;
    long double x = cos(u) + elevation * cos(lat_rad) / 6378140.0L;

  //mikroC has a bug in atan2: the X and Y arguments have to be reversed
  //This is the original code
  //delta_alpha_rad = atan2(-x * sin(xi_rad) * sin(h_rad), cos(delta_rad) - x * sin(xi_rad) * cos(h_rad));
    
  //This is the modified code
    delta_alpha_rad = atan2(cos(delta_rad) - x * sin(xi_rad) * cos(h_rad), - x * sin(xi_rad) * sin(h_rad));

  //mikroC has a bug in atan2: the X and Y arguments have to be reversed
  //This is the original code
  //*delta_prime = rad2deg(atan2((sin(delta_rad) - y * sin(xi_rad)) * cos(delta_alpha_rad), cos(delta_rad) - x * sin(xi_rad) * cos(h_rad)));
    
  //This is the modified code
    *delta_prime = rad2deg(atan2(cos(delta_rad) - x * sin(xi_rad) * cos(h_rad),(sin(delta_rad) - y * sin(xi_rad)) * cos(delta_alpha_rad)));

    *delta_alpha = rad2deg(delta_alpha_rad);
}
/*****************************************************************************************************************************************/


long double topocentric_right_ascension(long double alpha_deg, long double delta_alpha)
{
    return alpha_deg + delta_alpha;
}

long double topocentric_local_hour_angle(long double h, long double delta_alpha)
{
    return h - delta_alpha;
}

long double topocentric_elevation_angle(long double latitude, long double delta_prime, long double h_prime)
{
    long double lat_rad = deg2rad(latitude);
    long double delta_prime_rad = deg2rad(delta_prime);

    return rad2deg(asin(sin(lat_rad) * sin(delta_prime_rad) +
                        cos(lat_rad) * cos(delta_prime_rad) * cos(deg2rad(h_prime))));
}

long double atmospheric_refraction_correction(long double pressure, long double temperature, long double atmos_refract, long double e0)
{
    long double del_e = 0;

    if (e0 >= -1 * (SUN_RADIUS + atmos_refract))
        del_e = (pressure / 1010.0L) * (283.0L / (273.0L + temperature)) *
                 1.02L / (60.0L * tan(deg2rad(e0 + 10.3L / (e0 + 5.11L))));

    return del_e;
}

long double topocentric_elevation_angle_corrected(long double e0, long double delta_e)
{
    return e0 + delta_e;
}

long double topocentric_zenith_angle(long double e)
{
    return 90.0L - e;
}

long double topocentric_azimuth_angle_astro(long double h_prime, long double latitude, long double delta_prime){

    long double h_prime_rad = deg2rad(h_prime);
    long double lat_rad     = deg2rad(latitude);

    //char FormatBuf[64];
    //char* BufNum;


    long double Temp1;
    long double Temp2;
    long double Temp3;
    long double Temp4;
    long double Temp5;
/*
    BufNum = GetBuffer();
    LongDoubleToStr(h_prime, FormatBuf);
    sprintf(BufNum,"h_prime: ");
    strcat(BufNum,FormatBuf);
    USB_PutQueue(BufNum);

    BufNum = GetBuffer();
    LongDoubleToStr(latitude, FormatBuf);
    sprintf(BufNum,"latitude: ");
    strcat(BufNum,FormatBuf);
    USB_PutQueue(BufNum);

    BufNum = GetBuffer();
    LongDoubleToStr(delta_prime, FormatBuf);
    sprintf(BufNum,"delta_prime: ");
    strcat(BufNum,FormatBuf);
    USB_PutQueue(BufNum);
*/
    Temp1 = sin(h_prime_rad);
    Temp2 = cos(h_prime_rad) * sin(lat_rad) - tan(deg2rad(delta_prime)) * cos(lat_rad);
    Temp3 = atan2(Temp2,Temp1);
/*
    BufNum = GetBuffer();
    LongDoubleToStr(Temp3, FormatBuf);
    sprintf(BufNum,"Temp3: ");
    strcat(BufNum,FormatBuf);
    USB_PutQueue(BufNum);
*/
    Temp4 = rad2deg(Temp3);
/*
    BufNum = GetBuffer();
    LongDoubleToStr(Temp4, FormatBuf);
    sprintf(BufNum,"Temp4: ");
    strcat(BufNum,FormatBuf);
    USB_PutQueue(BufNum);
*/
    Temp5 = limit_degrees(Temp4);
/*
    BufNum = GetBuffer();
    LongDoubleToStr(Temp5, FormatBuf);
    sprintf(BufNum,"Temp5: ");
    strcat(BufNum,FormatBuf);
    USB_PutQueue(BufNum);
*/


  //mikroC has a bug in atan2: the X and Y arguments have to be reversed.
  //This is the original code.
  //return limit_degrees(rad2deg(atan2(sin(h_prime_rad), cos(h_prime_rad) * sin(lat_rad) - tan(deg2rad(delta_prime)) * cos(lat_rad))));

  //This calculation does not produce desired results by mikroC compiler. See test case at the bottom of this file.
  //return limit_degrees(rad2deg(atan2((cos(h_prime_rad) * sin(lat_rad)) - (tan(deg2rad(delta_prime)) * cos(lat_rad)), sin(h_prime_rad))));
  
  //Breaking the code into pieces seems to work
    return Temp5;
}

long double topocentric_azimuth_angle(long double azimuth_astro)
{
    return limit_degrees(azimuth_astro + 180.0L);
}

long double surface_incidence_angle(long double zenith, long double azimuth_astro, long double azm_rotation, long double slope)
{
    long double zenith_rad = deg2rad(zenith);
    long double slope_rad  = deg2rad(slope);

    return rad2deg(acos(cos(zenith_rad) * cos(slope_rad) + sin(slope_rad ) * sin(zenith_rad) *
                   cos(deg2rad(azimuth_astro - azm_rotation))));
}

long double sun_mean_longitude(long double jme)
{
    return limit_degrees(280.4664567L + jme * (360007.6982779L + jme * (0.03032028L +
                 jme * (1.0L/49931.0L   + jme * (-1.0L/15300.0L + jme * (-1.0L/2000000.0L))))));
}

long double eot(long double m, long double alpha, long double del_psi, long double epsilon)
{
    return limit_minutes(4.0L * (m - 0.0057183L - alpha + del_psi * cos(deg2rad(epsilon))));
}

long double approx_sun_transit_time(long double alpha_zero, long double longitude, long double nu)
{
    return (alpha_zero - longitude - nu) / 360.0L;
}

long double sun_hour_angle_at_rise_set(long double latitude, long double delta_zero, long double h0_prime)
{
    long double h0 = -99999.0L;
    long double latitude_rad = deg2rad(latitude);
    long double delta_zero_rad = deg2rad(delta_zero);
    long double argument = (sin(deg2rad(h0_prime)) - sin(latitude_rad)*sin(delta_zero_rad)) /
                           (cos(latitude_rad)*cos(delta_zero_rad));

    if (fabs(argument) <= 1) h0 = limit_degrees180(rad2deg(acos(argument)));

    return h0;
}

void approx_sun_rise_and_set(long double *m_rts, long double h0)
{
    long double h0_dfrac = h0 / 360.0L;

    m_rts[SUN_RISE]    = limit_zero2one(m_rts[SUN_TRANSIT] - h0_dfrac);
    m_rts[SUN_SET]     = limit_zero2one(m_rts[SUN_TRANSIT] + h0_dfrac);
    m_rts[SUN_TRANSIT] = limit_zero2one(m_rts[SUN_TRANSIT]);
}

long double rts_alpha_delta_prime(long double *ad, long double n)
{
    long double a = ad[JD_ZERO] - ad[JD_MINUS];
    long double b = ad[JD_PLUS] - ad[JD_ZERO];

    if (fabs(a) >= 2.0) a = limit_zero2one(a);
    if (fabs(b) >= 2.0) b = limit_zero2one(b);

    return ad[JD_ZERO] + n * (a + b + (b-a) * n) / 2.0L;
}

long double rts_sun_altitude(long double latitude, long double delta_prime, long double h_prime)
{
    long double latitude_rad = deg2rad(latitude);
    long double delta_prime_rad = deg2rad(delta_prime);

    return rad2deg(asin(sin(latitude_rad) * sin(delta_prime_rad) +
                        cos(latitude_rad) * cos(delta_prime_rad) * cos(deg2rad(h_prime))));
}

long double sun_rise_and_set(long double *m_rts, long double *h_rts, long double *delta_prime, long double latitude,
                             long double *h_prime, long double h0_prime, int sun)
{
    return m_rts[sun] + (h_rts[sun] - h0_prime) / (360.0L * cos(deg2rad(delta_prime[sun])) *
                     cos(deg2rad(latitude)) * sin(deg2rad(h_prime[sun])));
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Calculate required SPA parameters to get the right ascension (alpha) and declination (delta)
// Note: JD must be already calculated and in structure
///////////////////////////////////////////////////////////////////////////////////////////////////
void calculate_geocentric_sun_right_ascension_and_declination(spa_data *spa){

    long double x[TERM_X_COUNT];

    //char FormatBuf[64];
    //char* BufNum;

    spa->jc = julian_century(spa->jd);

    spa->jde = julian_ephemeris_day(spa->jd, spa->delta_t);
    spa->jce = julian_ephemeris_century(spa->jde);
    spa->jme = julian_ephemeris_millennium(spa->jce);

    spa->l = earth_heliocentric_longitude(spa->jme);
    spa->b = earth_heliocentric_latitude(spa->jme);
    spa->r = earth_radius_vector(spa->jme);

    spa->theta = geocentric_longitude(spa->l);
    spa->beta  = geocentric_latitude(spa->b);

    x[TERM_X0] = spa->x0 = mean_elongation_moon_sun(spa->jce);
    x[TERM_X1] = spa->x1 = mean_anomaly_sun(spa->jce);
    x[TERM_X2] = spa->x2 = mean_anomaly_moon(spa->jce);
    x[TERM_X3] = spa->x3 = argument_latitude_moon(spa->jce);
    x[TERM_X4] = spa->x4 = ascending_longitude_moon(spa->jce);

    nutation_longitude_and_obliquity(spa->jce, x, &(spa->del_psi), &(spa->del_epsilon));

    spa->epsilon0 = ecliptic_mean_obliquity(spa->jme);
    spa->epsilon  = ecliptic_true_obliquity(spa->del_epsilon, spa->epsilon0);

    spa->del_tau   = aberration_correction(spa->r);
    spa->lamda     = apparent_sun_longitude(spa->theta, spa->del_psi, spa->del_tau);
    spa->nu0       = greenwich_mean_sidereal_time (spa->jd, spa->jc);
    spa->nu        = greenwich_sidereal_time (spa->nu0, spa->del_psi, spa->epsilon);

    spa->alpha = geocentric_right_ascension(spa->lamda, spa->epsilon, spa->beta);

    spa->delta = geocentric_declination(spa->beta, spa->epsilon, spa->lamda);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Calculate Equation of Time (EOT) and Sun Rise, Transit, & Set (RTS)
///////////////////////////////////////////////////////////////////////////////////////////////////

void calculate_eot_and_sun_rise_transit_set(spa_data *spa)
{
    spa_data sun_rts;
    long double nu, m, h0, n;
    long double alpha[JD_COUNT], delta[JD_COUNT];
    long double m_rts[SUN_COUNT], nu_rts[SUN_COUNT], h_rts[SUN_COUNT];
    long double alpha_prime[SUN_COUNT], delta_prime[SUN_COUNT], h_prime[SUN_COUNT];
    long double h0_prime = -1*(SUN_RADIUS + spa->atmos_refract);
    int i;

    sun_rts = *spa;
    m = sun_mean_longitude(spa->jme);
    spa->eot = eot(m, spa->alpha, spa->del_psi, spa->epsilon);

    sun_rts.hour = sun_rts.minute = sun_rts.second = 0;
    sun_rts.delta_ut1 = sun_rts.timezone = 0;

    sun_rts.jd = julian_day (sun_rts.year, sun_rts.month, sun_rts.day, sun_rts.hour,
                             sun_rts.minute, sun_rts.second, sun_rts.delta_ut1, sun_rts.timezone);

    calculate_geocentric_sun_right_ascension_and_declination(&sun_rts);
    nu = sun_rts.nu;

    sun_rts.delta_t = 0;
    //sun_rts.jd--;  // Operator '--' is not applicable to these operands - mikroC compiler
    sun_rts.jd -= 1; // Replaced by this
    for (i = 0; i < JD_COUNT; i++) {
        calculate_geocentric_sun_right_ascension_and_declination(&sun_rts);
        alpha[i] = sun_rts.alpha;
        delta[i] = sun_rts.delta;
        //sun_rts.jd++;  // Operator '++' is not applicable to these operands - mikroC compiler
        sun_rts.jd += 1; // Replaced by this
    }

    m_rts[SUN_TRANSIT] = approx_sun_transit_time(alpha[JD_ZERO], spa->longitude, nu);
    h0 = sun_hour_angle_at_rise_set(spa->latitude, delta[JD_ZERO], h0_prime);

    if (h0 >= 0) {

        approx_sun_rise_and_set(m_rts, h0);

        for (i = 0; i < SUN_COUNT; i++) {

            nu_rts[i] = nu + 360.985647L*m_rts[i];

            n = m_rts[i] + spa->delta_t / 86400.0L;
            alpha_prime[i] = rts_alpha_delta_prime(alpha, n);
            delta_prime[i] = rts_alpha_delta_prime(delta, n);

            h_prime[i] = limit_degrees180pm(nu_rts[i] + spa->longitude - alpha_prime[i]);

            h_rts[i] = rts_sun_altitude(spa->latitude, delta_prime[i], h_prime[i]);
        }

        spa->srha = h_prime[SUN_RISE];
        spa->ssha = h_prime[SUN_SET];
        spa->sta  = h_rts[SUN_TRANSIT];

        spa->suntransit = dayfrac_to_local_hr(m_rts[SUN_TRANSIT] - h_prime[SUN_TRANSIT] / 360.0L,
                          spa->timezone);

        spa->sunrise = dayfrac_to_local_hr(sun_rise_and_set(m_rts, h_rts, delta_prime,
                       spa->latitude, h_prime, h0_prime, SUN_RISE), spa->timezone);

        spa->sunset  = dayfrac_to_local_hr(sun_rise_and_set(m_rts, h_rts, delta_prime,
                       spa->latitude, h_prime, h0_prime, SUN_SET),  spa->timezone);

    } else spa->srha= spa->ssha= spa->sta= spa->suntransit= spa->sunrise= spa->sunset= -99999.0L;

}


///////////////////////////////////////////////////////////////////////////////////////////////////
// Calculate all SPA parameters and put into structure
// Note: All inputs values (listed in header file) must already be in structure
///////////////////////////////////////////////////////////////////////////////////////////////////
int spa_calculate(spa_data *spa)
{
    int result;

    //char FormatBuf[64];
    //char* BufNum;

    result = validate_inputs(spa);

    if (result == 0)
    {
        spa->jd = julian_day (spa->year, spa->month, spa->day, spa->hour, spa->minute, spa->second,
                              spa->delta_ut1, spa->timezone);

        calculate_geocentric_sun_right_ascension_and_declination(spa);

        spa->h  = observer_hour_angle(spa->nu, spa->longitude, spa->alpha);
/*
        BufNum = GetBuffer();
        LongDoubleToStr(spa->h, FormatBuf);
        sprintf(BufNum,"h: ");
        strcat(BufNum,FormatBuf);
        USB_PutQueue(BufNum);
*/
        spa->xi = sun_equatorial_horizontal_parallax(spa->r);

        right_ascension_parallax_and_topocentric_dec(spa->latitude, spa->elevation, spa->xi, spa->h, spa->delta, &(spa->del_alpha),
                                                     &(spa->delta_prime));

        spa->alpha_prime = topocentric_right_ascension(spa->alpha, spa->del_alpha);
        spa->h_prime = topocentric_local_hour_angle(spa->h, spa->del_alpha);

        spa->e0 = topocentric_elevation_angle(spa->latitude, spa->delta_prime, spa->h_prime);
        spa->del_e = atmospheric_refraction_correction(spa->pressure, spa->temperature, spa->atmos_refract, spa->e0);

        spa->e = topocentric_elevation_angle_corrected(spa->e0, spa->del_e);

        spa->zenith = topocentric_zenith_angle(spa->e);

        spa->azimuth_astro = topocentric_azimuth_angle_astro(spa->h_prime, spa->latitude, spa->delta_prime);
/*
        BufNum = GetBuffer();
        LongDoubleToStr(spa->azimuth_astro, FormatBuf);
        sprintf(BufNum,"azimuth_astro: ");
        strcat(BufNum,FormatBuf);
        USB_PutQueue(BufNum);
*/

        spa->azimuth = topocentric_azimuth_angle(spa->azimuth_astro);

        if ((spa->function == SPA_ZA_INC) || (spa->function == SPA_ALL))
            spa->incidence  = surface_incidence_angle(spa->zenith, spa->azimuth_astro, spa->azm_rotation, spa->slope);

        if ((spa->function == SPA_ZA_RTS) || (spa->function == SPA_ALL))
            calculate_eot_and_sun_rise_transit_set(spa);

    }

    return result;
}
/*****************************************************************************************************************************************/

// Test that demonstrates wrong calculation in mikroC. Send command c to invoke.
void MyTest2(void){
     char FormatBuf[64];
     char* BufNum;

     long double h_prime = 11.10493181378699L;
     long double latitude = 39.74247741699219L;
     long double delta_prime = -9.316696485469796L;

     long double output1;

     long double Temp1;
     long double Temp2;
     long double Temp3;
     long double Temp4;
     long double Temp5;

     long double h_prime_rad = deg2rad(h_prime);
     long double lat_rad     = deg2rad(latitude);

     // This calculation gets a wrong result
     output1 = limit_degrees(rad2deg(atan2(cos(h_prime_rad) * sin(lat_rad) - tan(deg2rad(delta_prime)) * cos(lat_rad), sin(h_prime_rad))));

     BufNum = GetBuffer();
     LongDoubleToStr(output1, FormatBuf);
     sprintf(BufNum,"Wrong result: ");
     strcat(BufNum,FormatBuf);
     strcat(BufNum,"\r\n");
     USB_PutQueue(BufNum);

     // This calculation gets the correct result
     Temp1 = sin(h_prime_rad);
     Temp2 = cos(h_prime_rad) * sin(lat_rad) - tan(deg2rad(delta_prime)) * cos(lat_rad);
     Temp3 = atan2(Temp2,Temp1);  // mikroC transposes the X and Y arguments
/*
     BufNum = GetBuffer();
     LongDoubleToStr(Temp3, FormatBuf);
     sprintf(BufNum,"Temp3: ");
     strcat(BufNum,FormatBuf);
     USB_PutQueue(BufNum);
*/
     Temp4 = rad2deg(Temp3);
/*
     BufNum = GetBuffer();
     LongDoubleToStr(Temp4, FormatBuf);
     sprintf(BufNum,"Temp4: ");
     strcat(BufNum,FormatBuf);
     USB_PutQueue(BufNum);
*/
     Temp5 = limit_degrees(Temp4);

     BufNum = GetBuffer();
     LongDoubleToStr(Temp5, FormatBuf);
     sprintf(BufNum,"Right result: ");
     strcat(BufNum,FormatBuf);
     strcat(BufNum,"\r\n");
     USB_PutQueue(BufNum);
}