

/////////////////////////////////////////////
//          HEADER FILE for SPA.C          //
//                                         //
//      Solar Position Algorithm (SPA)     //
//                   for                   //
//        Solar Radiation Application      //
//                                         //
//               May 12, 2003              //
//                                         //
//   Filename: SPA.H                       //
//                                         //
//   Afshin Michael Andreas                //
//   afshin_andreas@nrel.gov (303)384-6383 //
//                                         //
//   Measurement & Instrumentation Team    //
//   Solar Radiation Research Laboratory   //
//   National Renewable Energy Laboratory  //
//   1617 Cole Blvd, Golden, CO 80401      //
/////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//                                                                    //
// Usage:                                                             //
//                                                                    //
//   1) In calling program, include this header file,                 //
//      by adding this line to the top of file:                       //
//           #include "spa.h"                                         //
//                                                                    //
//   2) In calling program, declare the SPA structure:                //
//           spa_data spa;                                            //
//                                                                    //
//   3) Enter the required input values into SPA structure            //
//      (input values listed in comments below)                       //
//                                                                    //
//   4) Call the SPA calculate function and pass the SPA structure    //
//      (prototype is declared at the end of this header file):       //
//           spa_calculate(&spa);                                     //
//                                                                    //
//   Selected output values (listed in comments below) will be        //
//   computed and returned in the passed SPA structure.  Output       //
//   will based on function code selected from enumeration below.     //
//                                                                    //
//   Note: A non-zero return code from spa_calculate() indicates that //
//         one of the input values did not pass simple bounds tests.  //
//         The valid input ranges and return error codes are also     //
//         listed below.                                              //
//                                                                    //
////////////////////////////////////////////////////////////////////////

#ifndef __solar_position_algorithm_header
#define __solar_position_algorithm_header








//enumeration for function codes to select desired final outputs from SPA
enum {
     SPA_ZA,           //calculate zenith and azimuth
     SPA_ZA_INC,       //calculate zenith, azimuth, and incidence
     SPA_ZA_RTS,       //calculate zenith, azimuth, and sun rise/transit/set values
     SPA_ALL,          //calculate all SPA output values
};

typedef struct
{
     //-------------------------INPUT VALUES-------------------------------------------------------

     int year;                  // 4-digit year,      valid range: -2000 to 6000, error code: 1
     int month;                 // 2-digit month,         valid range: 1 to  12,  error code: 2
     int day;                   // 2-digit day,           valid range: 1 to  31,  error code: 3
     int hour;                  // Observer local hour,   valid range: 0 to  24,  error code: 4
     int minute;                // Observer local minute, valid range: 0 to  59,  error code: 5
     long double second;        // Observer local second, valid range: 0 to <60,  error code: 6

     long double delta_ut1;     // Fractional second difference between UTC and UT which is used
                                // to adjust UTC for earth's irregular rotation rate and is derived
                                // from observation only and is reported in this bulletin:
                                // http://maia.usno.navy.mil/ser7/ser7.dat,
                                // where delta_ut1 = DUT1
                                // valid range: -1 to 1 second (exclusive), error code 17

     long double delta_t;       // Difference between earth rotation time and terrestrial time
                                // It is derived from observation only and is reported in this
                                // bulletin: http://maia.usno.navy.mil/ser7/ser7.dat,
                                // where delta_t = 32.184 + (TAI-UTC) - DUT1
                                // valid range: -8000 to 8000 seconds, error code: 7

     long double timezone;      // Observer time zone (negative west of Greenwich)
                                // valid range: -18   to   18 hours,   error code: 8

     long double longitude;     // Observer longitude (negative west of Greenwich)
                                // valid range: -180  to  180 degrees, error code: 9

     long double latitude;      // Observer latitude (negative south of equator)
                                // valid range: -90   to   90 degrees, error code: 10

     long double elevation;     // Observer elevation [meters]
                                // valid range: -6500000 or higher meters,    error code: 11

     long double pressure;      // Annual average local pressure [millibars]
                                // valid range:    0 to 5000 millibars,       error code: 12

     long double temperature;   // Annual average local temperature [degrees Celsius]
                                // valid range: -273 to 6000 degrees Celsius, error code; 13

     long double slope;         // Surface slope (measured from the horizontal plane)
                                // valid range: -360 to 360 degrees, error code: 14

     long double azm_rotation;  // Surface azimuth rotation (measured from south to projection of
                                // surface normal on horizontal plane, negative east)
                                // valid range: -360 to 360 degrees, error code: 15

     long double atmos_refract; // Atmospheric refraction at sunrise and sunset (0.5667 deg is typical)
                                // valid range: -5   to   5 degrees, error code: 16

     int function;              // Switch to choose functions for desired output (from enumeration)

     //-------------------------Intermediate OUTPUT VALUES-----------------------------------------

     long double jd;            //Julian day
     long double jc;            //Julian century

     long double jde;           //Julian ephemeris day
     long double jce;           //Julian ephemeris century
     long double jme;           //Julian ephemeris millennium

     long double l;             //earth heliocentric longitude [degrees]
     long double b;             //earth heliocentric latitude [degrees]
     long double r;             //earth radius vector [Astronomical Units, AU]

     long double theta;         //geocentric longitude [degrees]
     long double beta;          //geocentric latitude [degrees]

     long double x0;            //mean elongation (moon-sun) [degrees]
     long double x1;            //mean anomaly (sun) [degrees]
     long double x2;            //mean anomaly (moon) [degrees]
     long double x3;            //argument latitude (moon) [degrees]
     long double x4;            //ascending longitude (moon) [degrees]

     long double del_psi;       //nutation longitude [degrees]
     long double del_epsilon;   //nutation obliquity [degrees]
     long double epsilon0;      //ecliptic mean obliquity [arc seconds]
     long double epsilon;       //ecliptic true obliquity  [degrees]

     long double del_tau;       //aberration correction [degrees]
     long double lamda;         //apparent sun longitude [degrees]
     long double nu0;           //Greenwich mean sidereal time [degrees]
     long double nu;            //Greenwich sidereal time [degrees]

     long double alpha;         //geocentric sun right ascension [degrees]
     long double delta;         //geocentric sun declination [degrees]

     long double h;             //observer hour angle [degrees]
     long double xi;            //sun equatorial horizontal parallax [degrees]
     long double del_alpha;     //sun right ascension parallax [degrees]
     long double delta_prime;   //topocentric sun declination [degrees]
     long double alpha_prime;   //topocentric sun right ascension [degrees]
     long double h_prime;       //topocentric local hour angle [degrees]

     long double e0;            //topocentric elevation angle (uncorrected) [degrees]
     long double del_e;         //atmospheric refraction correction [degrees]
     long double e;             //topocentric elevation angle (corrected) [degrees]

     long double eot;           //equation of time [minutes]
     long double srha;          //sunrise hour angle [degrees]
     long double ssha;          //sunset hour angle [degrees]
     long double sta;           //sun transit altitude [degrees]

     //-------------------------Final OUTPUT VALUES------------------------------------------------

     long double zenith;        //topocentric zenith angle [degrees]
     long double azimuth_astro; //topocentric azimuth angle (westward from south) [for astronomers]
     long double azimuth;       //topocentric azimuth angle (eastward from north) [for navigators and solar radiation]
     long double incidence;     //surface incidence angle [degrees]

     long double suntransit;    //local sun transit time (or solar noon) [fractional hour]
     long double sunrise;       //local sunrise time (+/- 30 seconds) [fractional hour]
     long double sunset;        //local sunset time (+/- 30 seconds) [fractional hour]

} spa_data;

//-------------- Utility functions for other applications (such as NREL's SAMPA) ------------------

long double deg2rad(long double degrees);
long double rad2deg(long double radians);
long double deg2rad(long double degrees);
int integer(long double value);
long double limit_degrees(long double degrees);
long double limit_degrees180pm(long double degrees);
long double limit_degrees180(long double degrees);
long double limit_zero2one(long double value);
long double limit_minutes(long double minutes);
long double dayfrac_to_local_hr(long double dayfrac, long double timezone);
long double third_order_polynomial(long double a, long double b, long double c, long double d, long double x);

int validate_inputs(spa_data *spa);
long double julian_day (int year, int month, int day, int hour, int minute, long double second,
                        long double dut1, long double tz);

long double julian_century(long double jd);
long double julian_ephemeris_day(long double jd, long double delta_t);
long double julian_ephemeris_century(long double jde);
long double julian_ephemeris_millennium(long double jce);
//long double earth_periodic_term_summation(const long double terms[][TERM_COUNT], int count, long double jme);
long double earth_periodic_term_summation(const long double terms[][3], int count, long double jme);


long double earth_values(long double term_sum[], int count, long double jme);
long double earth_heliocentric_longitude(long double jme);
long double earth_heliocentric_latitude(long double jme);
long double earth_radius_vector(long double jme);
long double geocentric_longitude(long double l);
long double geocentric_latitude(long double b);
long double mean_elongation_moon_sun(long double jce);
long double mean_anomaly_sun(long double jce);
long double mean_anomaly_moon(long double jce);
long double argument_latitude_moon(long double jce);
long double ascending_longitude_moon(long double jce);
//long double xy_term_summation(int i, long double x[TERM_X_COUNT]);
//long double xy_term_summation(int i, long double x[5]);
//void nutation_longitude_and_obliquity(long double jce, long double x[TERM_X_COUNT], long double *del_psi, long double *del_epsilon);
void nutation_longitude_and_obliquity(long double jce, long double x[5], long double *del_psi, long double *del_epsilon);
long double ecliptic_mean_obliquity(long double jme);
long double ecliptic_true_obliquity(long double delta_epsilon, long double epsilon0);
long double aberration_correction(long double r);
long double apparent_sun_longitude(long double theta, long double delta_psi, long double delta_tau);
long double greenwich_mean_sidereal_time (long double jd, long double jc);
long double greenwich_sidereal_time (long double nu0, long double delta_psi, long double epsilon);
long double geocentric_right_ascension(long double lamda, long double epsilon, long double beta);
long double geocentric_declination(long double beta, long double epsilon, long double lamda);
long double observer_hour_angle(long double nu, long double longitude, long double alpha_deg);
long double sun_equatorial_horizontal_parallax(long double r);
void right_ascension_parallax_and_topocentric_dec(long double latitude, long double elevation, long double xi, long double h,
                                                  long double delta, long double *delta_alpha, long double *delta_prime);
long double topocentric_right_ascension(long double alpha_deg, long double delta_alpha);
long double topocentric_local_hour_angle(long double h, long double delta_alpha);
long double topocentric_elevation_angle(long double latitude, long double delta_prime, long double h_prime);
long double atmospheric_refraction_correction(long double pressure, long double temperature, long double atmos_refract, long double e0);
long double topocentric_elevation_angle_corrected(long double e0, long double delta_e);
long double topocentric_zenith_angle(long double e);
long double topocentric_azimuth_angle_astro(long double h_prime, long double latitude, long double delta_prime);
long double topocentric_azimuth_angle(long double azimuth_astro);
long double sun_mean_longitude(long double jme);
long double eot(long double m, long double alpha, long double del_psi, long double epsilon);
long double approx_sun_transit_time(long double alpha_zero, long double longitude, long double nu);
long double sun_hour_angle_at_rise_set(long double latitude, long double delta_zero, long double h0_prime);
void approx_sun_rise_and_set(long double *m_rts, long double h0);
long double rts_alpha_delta_prime(long double *ad, long double n);
long double rts_sun_altitude(long double latitude, long double delta_prime, long double h_prime);
long double sun_rise_and_set(long double *m_rts, long double *h_rts, long double *delta_prime, long double latitude, long double *h_prime,
                             long double h0_prime, int sun);

void calculate_geocentric_sun_right_ascension_and_declination(spa_data *spa);
void calculate_eot_and_sun_rise_transit_set(spa_data *spa);



int spa_calculate(spa_data *spa);



/* spa_tester functions */
// Enter initial data and do calculations
void SPA_Test(void);

// Test that demonstrates wrong calculation in mikroC
void MyTest2(void);



#endif